"""
Emit parameters to pip extracted from a script.
"""

import pathlib

import autocommand
import path

from .scripts import DepsReader


def separator(input) -> str:
    lookup = dict(space=' ', newline='\n', null='\0')
    return lookup.get(input, input)


@autocommand.autocommand(__name__)
def run(
    script: path.ExtantFile,
    separator: separator = ' ',  # type: ignore[valid-type]
):
    """
    >>> run(['examples/test-mongodb-covered-query.py'])
    pytest jaraco.mongodb
    >>> run(['does-not-exist'])
    Traceback (most recent call last):
    FileNotFoundError: does-not-exist does not exist as a file.
    >>> run(['examples/test-mongodb-covered-query.py', '--separator', 'newline'])
    pytest
    jaraco.mongodb
    """
    joiner = separator.join  # type: ignore[attr-defined]
    print(joiner(DepsReader.try_read(pathlib.Path(script)).params()))



import contextlib
import hashlib

import platformdirs

from .. import deps

paths = platformdirs.PlatformDirs(appname='pip run', appauthor=False)


class Hash:
    """
    Hash class with support for unicode text.
    """

    def __init__(self, name):
        self._hash = hashlib.new(name)

    def update(self, text):
        self._hash.update(text.encode('utf-8'))

    def hexdigest(self):
        return self._hash.hexdigest()


def cache_key(args):
    """
    Generate a cache key representing the packages to be installed.

    >>> reqs1, reqs2 = getfixture('reqs_files')
    >>> cache_key(['-r', str(reqs1), '--requirement', str(reqs2), 'requests'])
    '88d9f8a3a4009c1f685a7a724519bd5187e1227d72be6bc7f20a4a02f36d14b3'

    The key should be insensitive to order.

    >>> cache_key(['--requirement', str(reqs2), 'requests', '-r', str(reqs1)])
    '88d9f8a3a4009c1f685a7a724519bd5187e1227d72be6bc7f20a4a02f36d14b3'

    >>> cache_key(['--foo', '-q'])
    'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    """
    parsed = deps.Install.parse(args)
    hash = Hash('sha256')
    for req in sorted(parsed.package):
        hash.update(req + '\n')
    for file in sorted(parsed.requirement):
        hash.update('req:\n' + file.read_text(encoding='utf-8'))
    return hash.hexdigest()


@contextlib.contextmanager
def context(args):
    yield paths.user_cache_path.joinpath(cache_key(args))



import argparse
import contextlib
import pathlib

from jaraco import env  # type: ignore[attr-defined] # (python/mypy#10360)
from jaraco.context import suppress
from jaraco.functools import bypass_when
from more_itertools import locate, split_at

from .compat.py38 import files


def _is_python_arg(item: str):
    """
    Return True if the item can be inferred as a parameter
    to Python and not to pip install.
    """
    path = pathlib.Path(item)
    return path.is_file() and (path.suffix == '.py' or _has_shebang(path))


@suppress(UnicodeDecodeError)
def _has_shebang(path: pathlib.Path) -> bool:
    with path.open(encoding='utf-8-sig') as fp:
        first_line = fp.readline()
    return first_line.startswith("#!")


def _separate_script(args):
    """
    Split arguments into install and python args based on inferenece.

    >>> _separate_script(['foo', 'bar'])
    (['foo', 'bar'], [])
    >>> _separate_script(['foo', 'pip-run.py', 'bar'])
    (['foo'], ['pip-run.py', 'bar'])
    >>> _separate_script(['path.py', 'pip-run.py'])
    (['path.py'], ['pip-run.py'])
    >>> _separate_script(['path.py', 'pip-run.py', 'pip-run.py'])
    (['path.py'], ['pip-run.py', 'pip-run.py'])
    >>> _separate_script(['README.rst'])
    (['README.rst'], [])
    >>> _separate_script(['pip-run.py'])
    ([], ['pip-run.py'])
    >>> _separate_script(['pip-run.py', 'pip-run.py'])
    ([], ['pip-run.py', 'pip-run.py'])
    """
    pivot = next(locate(args, _is_python_arg), len(args))
    return args[:pivot], args[pivot:]


def _separate_dash(args):
    """
    Separate args based on a dash separator.

    >>> _separate_dash(['foo', '--', 'bar'])
    (['foo'], ['bar'])

    >>> _separate_dash(['foo', 'bar', '--'])
    (['foo', 'bar'], [])

    >>> _separate_dash(['foo', 'bar'])
    Traceback (most recent call last):
    ...
    ValueError: ...
    """
    pre, post = split_at(args, '--'.__eq__, maxsplit=1)
    return pre, post


def separate(args):
    """
    Separate the command line arguments into arguments for pip
    and arguments to Python.

    >>> separate(['foo', '--', 'bar'])
    (['foo'], ['bar'])
    >>> separate(['foo', 'bar'])
    (['foo', 'bar'], [])
    """
    with contextlib.suppress(ValueError):
        return _separate_dash(args)

    return _separate_script(args)


@bypass_when(env.Check('PIP_RUN_IPYTHON_MODE', default='infer', expect='ignore'))
def infer_ipython(sep_args):
    """
    Check for the presence of the argument 'ipython' in pip_args.

    If present and py_args is empty, return arguments for pip and
    Python that when run will start an ipython interpreter.

    >>> infer_ipython((['ipython', 'foo'], []))
    (['ipython', 'foo'], ['-m', 'IPython'])
    >>> infer_ipython((['ipython', 'foo'], ['bar']))
    (['ipython', 'foo'], ['bar'])
    >>> infer_ipython((['foo'], ['bar']))
    (['foo'], ['bar'])

    >>> getfixture('monkeypatch').setenv('PIP_RUN_IPYTHON_MODE', 'ignore')
    >>> infer_ipython((['ipython', 'foo'], []))
    (['ipython', 'foo'], [])
    >>> getfixture('monkeypatch').setenv('PIP_RUN_IPYTHON_MODE', 'infer')
    >>> infer_ipython((['ipython', 'foo'], []))
    (['ipython', 'foo'], ['-m', 'IPython'])
    """
    pip_args, py_args = sep_args
    use_ipython = not py_args and 'ipython' in pip_args

    return (
        pip_args,
        ['-m', 'IPython'] * use_ipython + py_args,
    )


def intercept(args):
    """
    Detect certain args and intercept them.
    """
    usage = files(__package__).joinpath('usage.txt').read_text(encoding='utf-8')
    argparse.ArgumentParser(usage=usage).parse_known_args(args)



import pytest

from pip_run import commands

valid_shebangs = [
    '#!/usr/bin/env python',
    '#!/usr/bin/env -S pip-run',
    '#!/usr/bin/python -W error',
    # with BOM
    '#!/usr/bin/env -S python'.encode('utf-8-sig').decode('utf-8'),
]

invalid_shebangs = [
    '#/usr/bin/env python',
    '!/usr/bin/env python',
]


@pytest.mark.parametrize('shebang', valid_shebangs)
def test_shebang_detected(tmp_path, shebang):
    script = tmp_path / 'script'
    script.write_text(f'{shebang}\nprint("Hello world!")', encoding='utf-8')
    assert commands._has_shebang(script)


@pytest.mark.parametrize('shebang', invalid_shebangs)
def test_shebang_not_detected(tmp_path, shebang):
    script = tmp_path / 'script'
    script.write_text(f'{shebang}\nprint("Hello world!")', encoding='utf-8')
    assert not commands._has_shebang(script)


def test_shebang_invalid_encoding(tmp_path):
    """
    If the script cannot be decoded in UTF-8, value should be false.
    """
    script = tmp_path / 'script'
    shebang = '\xf1#!/usr/bin/env -S python'
    script.write_text(f'{shebang}\nprint("Hello world!")', encoding='latin-1')
    assert not commands._has_shebang(script)



import argparse
import contextlib
import importlib
import os
import pathlib
import subprocess
import sys
import types

from jaraco.context import suppress

from .compat.py38 import subprocess_path as sp


class Install(types.SimpleNamespace):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-r',
        '--requirement',
        action='append',
        type=pathlib.Path,
        default=[],
    )
    parser.add_argument('package', nargs='*')

    @classmethod
    def parse(cls, args):
        parsed, unused = cls.parser.parse_known_args(args)
        return cls(**vars(parsed))

    def __bool__(self):
        """
        Return True only if the args to pip install
        indicate something to install.

        >>> bool(Install.parse(['inflect']))
        True
        >>> bool(Install.parse(['-q']))
        False
        >>> bool(Install.parse(['-q', 'inflect']))
        True
        >>> bool(Install.parse(['-rfoo.txt']))
        True
        >>> bool(Install.parse(['projects/inflect']))
        True
        >>> bool(Install.parse(['~/projects/inflect']))
        True
        """
        return bool(self.requirement or self.package)


def retention_strategy():
    strategy = os.environ.get('PIP_RUN_RETENTION_STRATEGY') or 'destroy'
    return importlib.import_module(f'.retention.{strategy}', package=__package__)


@suppress(FileNotFoundError)
def contents(path):
    return list(path.iterdir())


def empty(path):
    """
    >>> target = getfixture('tmp_path')
    >>> empty(target)
    True
    >>> _ = target.joinpath('file.txt').write_text('contents', encoding='utf-8')
    >>> empty(target)
    False

    A non-existent path is considered empty.

    >>> empty(target / 'missing')
    True
    """
    return not bool(contents(path))


@contextlib.contextmanager
def load(*args):
    with retention_strategy().context(args) as target:
        cmd = (sys.executable, '-m', 'pip', 'install', '-t', sp(target)) + args
        env = dict(os.environ, PIP_QUIET="1")
        if Install.parse(args) and empty(target):
            subprocess.check_call(cmd, env=env)
        yield target



import abc
import ast
import contextlib
import itertools
import json
import pathlib
import re

import packaging.requirements
from coherent.deps import imports, pypi
from jaraco.context import suppress
from jaraco.functools import compose

from .compat.py310 import tomllib

ValidRequirementString = compose(str, packaging.requirements.Requirement)

_unique = dict.fromkeys


class EllipsisFilter:
    found = False

    def __call__(self, item):
        is_ellipsis = item is Ellipsis
        self.found |= is_ellipsis
        return not is_ellipsis


class Dependencies(list):
    index_url = None

    def params(self):
        return ['--index-url', self.index_url] * bool(self.index_url) + self

    @classmethod
    def load(cls, items):
        """
        Construct self from items, validated as requirements.
        """
        ef = EllipsisFilter()
        deps = cls(map(ValidRequirementString, filter(ef, items)))
        deps.inferred = ef.found
        return deps


class DepsReader:
    """
    Given a Python script, read the dependencies it declares.
    Does not execute the script, so expects __requires__ to be
    assigned a static list of strings.
    """

    def __init__(self, script):
        self.script = script

    @classmethod
    def try_read(cls, script_path: pathlib.Path):
        results = (subclass._try_read(script_path) for subclass in cls.__subclasses__())
        return next(filter(None, results), Dependencies())

    @classmethod
    @suppress(Exception)
    def _try_read(cls, script_path: pathlib.Path):
        """
        Attempt to load the dependencies from the script,
        but return None if unsuccessful.
        """
        reader = cls.load(script_path)
        return reader.maybe_infer(reader.read())

    @classmethod
    @abc.abstractmethod
    def load(cls, script: pathlib.Path):
        """
        Construct a DepsReader from the script path.
        """

    @classmethod
    def search(cls, params):
        """
        Given a (possibly-empty) series of parameters to a
        Python interpreter, return any dependencies discovered
        in a script indicated in the parameters. Only honor the
        first file found.
        """
        safe_is_file = suppress(OSError)(pathlib.Path.is_file)
        files = filter(safe_is_file, map(pathlib.Path, params))
        return cls.try_read(next(files, None)).params()

    def maybe_infer(self, deps):
        if deps.inferred:
            deps[:] = _unique(itertools.chain(deps, self.infer()))
        return deps

    def infer(self):
        r"""
        >>> DepsReader('import sys\nimport cowsay\nimport jaraco.text.missing\n').infer()
        ['cowsay', 'jaraco.text']
        """
        return Dependencies(
            pypi.distribution_for(imp)
            for imp in imports.get_module_imports(self.script)
            if not imp.excluded()
        )

    def read(self):
        return self.read_toml() or self.read_python()

    def read_toml(self):
        r"""
        >>> DepsReader('# /// script\n# dependencies = ["foo", "bar"]\n# ///\n').read()
        ['foo', 'bar']
        >>> DepsReader('# /// pyproject\n# dependencies = ["foo", "bar"]\n# ///\n').read_toml()
        []
        >>> DepsReader('# /// pyproject\n#dependencies = ["foo", "bar"]\n# ///\n').read_toml()
        []
        >>> DepsReader('# /// script\n# dependencies = ["foo", "bar"]\n').read_toml()
        []
        >>> DepsReader('# /// script\n# ///\n\n# /// script\n# ///').read_toml()
        Traceback (most recent call last):
        ...
        ValueError: Multiple script blocks found
        """
        TOML_BLOCK_REGEX = (
            r'(?m)^# /// (?P<type>[a-zA-Z0-9-]+)$\s(?P<content>(^#(| .*)$\s)*)^# ///$'
        )
        name = 'script'
        matches = list(
            filter(
                lambda m: m.group('type') == name,
                re.finditer(TOML_BLOCK_REGEX, self.script),
            )
        )
        if len(matches) > 1:
            raise ValueError(f'Multiple {name} blocks found')
        elif len(matches) == 1:
            content = ''.join(
                line[2:] if line.startswith('# ') else line[1:]
                for line in matches[0].group('content').splitlines(keepends=True)
            )
            deps = tomllib.loads(content).get("dependencies", [])
        else:
            deps = []
        return Dependencies.load(deps)

    def read_python(self):
        r"""
        >>> DepsReader("__requires__=['foo']").read()
        ['foo']
        >>> empty = DepsReader("").read()
        >>> empty
        []
        >>> empty.inferred
        True
        """
        reqs = suppress(ValueError)(self._read)('__requires__')
        if reqs is None:
            reqs = [...]
        deps = Dependencies.load(reqs)
        with contextlib.suppress(ValueError):
            deps.index_url = self._read('__index_url__')
        return deps

    def _read(self, var_name):
        """
        Read a variable from self.script by parsing the AST.

        Raises ValueError if the variable is not found or if it
        appears more than once.
        """
        mod = ast.parse(self.script)
        (node,) = (
            node
            for node in mod.body
            if isinstance(node, ast.Assign)
            and len(node.targets) == 1
            and isinstance(node.targets[0], ast.Name)
            and node.targets[0].id == var_name
        )
        return ast.literal_eval(node.value)


class SourceDepsReader(DepsReader):
    @classmethod
    def load(cls, script: pathlib.Path):
        return cls(script.read_text(encoding='utf-8'))


def _load_json(path: pathlib.Path):
    with path.open(encoding='utf-8') as stream:
        return json.load(stream)


class NotebookDepsReader(DepsReader):
    @classmethod
    def load(cls, script: pathlib.Path):
        lines = (
            line
            for cell in _load_json(script)['cells']
            for line in cell['source'] + ['\n']
            if cell['cell_type'] == 'code' and not line.startswith('%')
        )
        return cls(''.join(lines))



import pytest

import pip_run.deps as deps


@pytest.mark.usefixtures('retention_strategy')
class TestLoad:
    def test_no_args_passes(self):
        """
        If called with no arguments, load() should still provide
        a context.
        """
        with deps.load():
            pass

    def test_only_options_passes(self):
        """
        If called with only options, but no installable targets,
        load() should still provide a context.
        """
        with deps.load('-q'):
            pass


@pytest.mark.usefixtures('retention_strategy')
def test_target_retention_context():
    """Verify a target exists or can be created."""
    with deps.retention_strategy().context([]) as target:
        target.mkdir(exist_ok=True)



import itertools
import os
import pathlib
import signal
import subprocess
import sys
import textwrap

from jaraco.context import suppress


def inject_sitecustomize(target: pathlib.Path):
    r"""
    Create a sitecustomize file in the target that will install
    the target as a sitedir.

    >>> tmp_path = getfixture('tmp_path')
    >>> inject_sitecustomize(tmp_path)
    >>> sc = tmp_path / 'sitecustomize.py'
    >>> 'Path' not in sc.read_text(encoding='utf-8')
    True
    """
    hook = textwrap.dedent(
        f"""
        import site
        site.addsitedir({os.fspath(target)!r})
        """
    ).lstrip()
    target.joinpath('sitecustomize.py').write_text(hook, encoding='utf-8')


_unique = dict.fromkeys


def _path_insert(previous, value):
    """
    Given a pathsep-separated env key, insert value.

    >>> orig = os.pathsep.join(['foo', 'bar'])
    >>> addl = _path_insert(orig, 'bing')
    >>> addl.split(os.pathsep)
    ['bing', 'foo', 'bar']
    >>> dupl = _path_insert(orig, 'bar')
    >>> dupl.split(os.pathsep)
    ['bar', 'foo']
    """
    prefix = (value,)
    suffix = filter(None, previous.split(os.pathsep))
    return os.pathsep.join(_unique(itertools.chain(prefix, suffix)))


def _build_env(target, *, orig=os.environ):
    """
    Prepend target to PYTHONPATH and add $target/bin to PATH.

    >>> orig = dict(PYTHONPATH='/orig', PATH='/orig')
    >>> env = _build_env(pathlib.Path('/tmp/pip-run/target'), orig=orig)
    >>> pprint(norm_env_paths(env))
    {'PATH': '/tmp/pip-run/target/bin:/orig',
     'PYTHONPATH': '/tmp/pip-run/target:/orig'}
    """
    overlay = dict(
        PYTHONPATH=_path_insert(orig.get('PYTHONPATH', ''), os.fspath(target)),
        PATH=_path_insert(orig.get('PATH', ''), os.fspath(target / 'bin')),
    )
    return {**orig, **overlay}


def _setup_env(target):
    inject_sitecustomize(target)
    return _build_env(target)


def _ensure_remove_prefix(text: str, prefix: str) -> str:
    start, found, rest = text.partition(prefix)
    if start or not found:
        raise ValueError(f"No prefix {prefix} for {text}")
    return rest


@suppress(IndexError, ValueError)
def _strip_bang(params):
    """
    Strip a literal `!` from the first parameter or return None.
    """
    cmd, rest = params[0], params[1:]
    return [_ensure_remove_prefix(cmd, '!')] + rest


def infer_cmd(params):
    """
    From params, infer command args for a subprocess.

    If the first parameter starts with a ``!``, it's a literal
    subcommand. Otherwise, it's parameters to a Python
    subprocess.

    >>> infer_cmd(['a.py', 'param1'])
    ['...', 'a.py', 'param1']
    >>> infer_cmd(['!an-exe', 'param1'])
    ['an-exe', 'param1']
    >>> infer_cmd([]) == [sys.executable]
    True
    """
    return _strip_bang(params) or [sys.executable] + params


def with_path(target: pathlib.Path, cmd):
    """
    Launch cmd with target on the path
    """

    def null_handler(signum, frame):
        pass  # pragma: no cover

    signal.signal(signal.SIGINT, null_handler)
    return subprocess.Popen(cmd, env=_setup_env(target)).wait()



import ntpath
import posixpath
import pprint
import textwrap

import jaraco.path
import pytest

import pip_run.retention.persist

collect_ignore = ['examples']


@pytest.fixture
def reqs_files(tmp_path):
    """Create a couple of requirements files."""
    jaraco.path.build(
        {
            'reqs1.txt': textwrap.dedent(
                """
                abc
                def
                """
            ).lstrip(),
            'reqs2.txt': textwrap.dedent(
                """
                uvw
                xyz
                """
            ).lstrip(),
        },
        tmp_path,
    )
    return tmp_path.glob('reqs*.txt')


@pytest.fixture(scope="session")
def monkeypatch_session():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(autouse=True, scope='session')
def alt_cache_dir(monkeypatch_session, tmp_path_factory):
    alt_cache = tmp_path_factory.mktemp('cache')

    class Paths:
        user_cache_path = alt_cache

    monkeypatch_session.setattr(pip_run.retention.persist, 'paths', Paths)


@pytest.fixture(params=['persist', 'destroy'])
def retention_strategy(monkeypatch, request):
    monkeypatch.setenv('PIP_RUN_RETENTION_STRATEGY', request.param)


@pytest.fixture
def doctest_namespace():
    def norm_path(path):
        return path.replace(ntpath.sep, posixpath.sep).replace(
            ntpath.pathsep, posixpath.pathsep
        )

    def norm_env_paths(env):
        return {key: norm_path(value) for key, value in env.items()}

    return dict(locals(), pprint=pprint.pprint)



"""
This script demonstrates how ``pip-run`` facilitates the
simple execution of complex tasks with their
dependencies.

Run this example with ``pip-run $script``.

It creates a MongoDB instance, and then runs some
assertions against it.

The script uses `jaraco.mongodb
<https://pypi.org/project/jaraco.mongodb>`_ to (maybe)
install and configure an ephemeral instance of MongoDB.

Running this script with ``pip-run``` leaves no trace of its
execution, other than adding packages to the pip
cache (if available), so feel free to give it a try.

This script gets read three times:

- First ``pip-run`` reads it statically to determine the
  dependencies found in ``__requires__`` and
  installs those dependencies in a temp dir.
- Then, the script is executed by a Python interpreter
  which executes the script up to the __main__ block,
  which invokes pytest with the script name as the
  first argument.
- ``py.test`` then runs the test, using the plugins
  from jaraco.mongodb to satisfy the ``mongodb_instance``
  fixture for the test.

Alternately, the script could be
invoked thus:  ``pip-run -- -m pytest $script``, which would
eliminate the need for the ``__main__`` block and
would skip the second processing of the script.
"""

if __name__ == '__main__':
    # invoke pytest on this script
    import sys

    import pytest

    sys.exit(pytest.main(['--noconftest'] + sys.argv))

import itertools
import random

import pytest
from jaraco.mongodb.testing import assert_covered


@pytest.fixture
def docs_in_db(mongodb_instance):
    """
    Install 100 records with random numbers
    """
    conn = mongodb_instance.get_connection()
    coll = conn.test_db.test_coll
    coll.drop()
    coll.create_index('number')
    n_records = 100
    for _ in itertools.islice(itertools.count(), n_records):
        doc = dict(number=random.randint(0, 2**32 - 1), value='some value')
        conn.test_db.test_coll.insert_one(doc)
    return coll


def test_records(docs_in_db, mongodb_instance):
    "Test 100 records are present and query is covered"
    # load all the numbers to ensure the index is in RAM
    _hint = 'number_1'
    _filter = {'number': {'$gt': 0}}
    _projection = {'number': True, '_id': False}
    cur = docs_in_db.find(filter=_filter, projection=_projection).hint(_hint)
    assert_covered(cur)

    # consume the cursor for good measure
    docs = list(cur)
    assert len(docs) == 100



import sys

from . import commands, deps, launch, scripts


def run(args=sys.argv[1:]):
    """
    Main entry point for pip-run.
    """
    pip_args, run_args = commands.infer_ipython(commands.separate(args))
    commands.intercept(pip_args)
    pip_args.extend(scripts.DepsReader.search(run_args))
    with deps.load(*pip_args) as home:
        raise SystemExit(launch.with_path(home, launch.infer_cmd(run_args)))



import platform
import sys

from jaraco.functools import identity

subprocess_path = (
    str if sys.version_info < (3, 9) and platform.system() == 'Windows' else identity
)

if sys.version_info >= (3, 9):
    from importlib.resources import files as files
else:  # pragma: no cover
    from importlib_resources import files as files



import itertools
import subprocess
import sys
import textwrap

import jaraco.path
import nbformat
import pytest

from pip_run import scripts


def DALS(str):
    return textwrap.dedent(str).lstrip()


@pytest.mark.network
@pytest.mark.parametrize(
    'shebang, extension, dash',
    itertools.product(('', '#!/usr/bin/env python'), ('', '.py'), ([], ['--'])),
)
def test_pkg_imported(tmp_path, shebang, extension, dash):
    """
    Create a script that loads a package and ensure it runs.
    """
    shebang or extension or dash or pytest.skip('Unable to infer script')
    scriptname = 'script' + extension
    jaraco.path.build(
        {
            scriptname: DALS(
                f"""
                {shebang}
                import sample
                print("Import succeeded")
                """
            )
        },
        tmp_path,
    )
    script = tmp_path / scriptname
    pip_args = ['sampleproject']
    script_args = dash + [str(script)]
    cmd = [sys.executable, '-m', 'pip-run'] + pip_args + script_args

    out = subprocess.check_output(cmd, text=True, encoding='utf-8')
    assert 'Import succeeded' in out


class TestSourceDepsReader:
    def test_reads_files_with_attribute_assignment(self):
        script = textwrap.dedent(
            """
            __requires__=['foo']
            x.a = 'bar'
            """
        )
        assert scripts.DepsReader(script).read() == ['foo']

    def test_duplicate_definition(self):
        script = textwrap.dedent(
            """
            __requires__=['foo']
            __requires__=['bar']
            """
        )
        assert scripts.DepsReader(script).read() == []

    def test_reads_files_with_multiple_assignment(self):
        script = textwrap.dedent(
            """
            __requires__=['foo']
            x, a = [a, x]
            """
        )
        assert scripts.DepsReader(script).read() == ['foo']

    def test_index_url(self):
        script = textwrap.dedent(
            """
            __requires__ = ['foo']
            __index_url__ = 'https://my.private.index/'
            """
        )
        reqs = scripts.DepsReader(script).read()
        assert reqs.index_url == 'https://my.private.index/'

    def test_toml_style(self):
        script = textwrap.dedent(
            """
            #! shebang

            # /// script
            # dependencies = ["foo == 3.1"]
            # ///
            """
        )
        reqs = scripts.DepsReader(script).read()
        assert reqs == ['foo==3.1']

    def test_search_long_parameter(self):
        """
        A parameter that is too long to be a filename should not fail.
        """
        exes = 'x' * 248
        command = f'print("{exes}")'
        params = ['-c', command]
        scripts.DepsReader.search(params)


class TestNotebookDepsReader:
    @pytest.fixture
    def notebook_factory(self, tmpdir, request):
        class Factory:
            def __init__(self):
                self.nb = nbformat.v4.new_notebook()
                self.path = tmpdir / (request.node.name + '.ipynb')

            @property
            def filename(self):
                return str(self.path)

            def write(self):
                nbformat.write(self.nb, self.filename)

            def add_code(self, code):
                self.nb['cells'].append(nbformat.v4.new_code_cell(code))

            def add_markdown(self, text):
                self.nb['cells'].append(nbformat.v4.new_markdown_cell(text))

        return Factory()

    def test_one_code_block(self, notebook_factory):
        notebook_factory.add_code('__requires__ = ["matplotlib"]')
        notebook_factory.write()
        reqs = scripts.DepsReader.try_read(notebook_factory.path)
        assert reqs == ['matplotlib']

    def test_multiple_code_blocks(self, notebook_factory):
        notebook_factory.add_code('__requires__ = ["matplotlib"]')
        notebook_factory.add_code("import matplotlib")
        notebook_factory.write()
        reqs = scripts.DepsReader.try_read(notebook_factory.path)
        assert reqs == ['matplotlib']

    def test_code_and_markdown(self, notebook_factory):
        notebook_factory.add_code('__requires__ = ["matplotlib"]')
        notebook_factory.add_markdown("Mark this down please")
        notebook_factory.write()
        reqs = scripts.DepsReader.try_read(notebook_factory.path)
        assert reqs == ['matplotlib']

    def test_jupyter_directives(self, notebook_factory):
        notebook_factory.add_code('__requires__ = ["matplotlib"]')
        notebook_factory.add_code("%matplotlib inline\nimport matplotlib")
        notebook_factory.write()
        reqs = scripts.DepsReader.try_read(notebook_factory.path)
        assert reqs == ['matplotlib']


@pytest.mark.network
def test_pkg_loaded_from_alternate_index(tmp_path):
    """
    Create a script that loads cython from an alternate index
    and ensure it runs.
    """
    jaraco.path.build(
        {
            'script': DALS(
                """
                __requires__ = ['sampleproject']
                __index_url__ = 'https://devpi.net/root/pypi/+simple/'
                import sample
                print("Import succeeded")
                """
            )
        },
        tmp_path,
    )
    cmd = [sys.executable, '-m', 'pip-run', '-v', '--', str(tmp_path / 'script')]

    out = subprocess.check_output(cmd, text=True, encoding='utf-8')
    assert 'Import succeeded' in out
    assert 'devpi.net' in out


def _minimal_flit(name):
    return {
        'pyproject.toml': DALS(
            f"""
            [project]
            name = "{name}"
            # flit requires a version
            version = "1"
            # flit requires a description
            description = ""

            [build-system]
            build-backend = "flit_core.buildapi"
            requires = ["flit_core"]
            """
        ),
        name: {
            # flit requires an init file
            '__init__.py': '',
        },
    }


def test_pkg_loaded_from_url(tmp_path):
    """
    Create a script whose dependency is only installable
    from a custom url and ensure it runs.
    """
    dependency = tmp_path / 'barbazquux-1.0'
    url_req = f'barbazquux @ file://{dependency.as_posix()}'
    jaraco.path.build(
        {
            'barbazquux-1.0': _minimal_flit('barbazquux'),
            'script_dir': {
                'script': DALS(
                    f"""
                    __requires__ = [{url_req!r}]
                    import barbazquux
                    print("Successfully imported barbazquux")
                    """
                ),
            },
        },
        tmp_path,
    )

    script = tmp_path.joinpath('script_dir', 'script')
    cmd = [
        sys.executable,
        '-m',
        'pip-run',
        '--no-index',
        '--no-build-isolation',
        '--',
        str(script),
    ]
    out = subprocess.check_output(cmd, text=True, encoding='utf-8')
    assert 'Successfully imported barbazquux' in out



import os
import pathlib

import pytest

from pip_run import launch


def test_with_path(tmp_path, capfd):
    params = ['-c', 'import sys; sys.stdout.write("\\n".join(sys.path))']
    res = launch.with_path(tmp_path, launch.infer_cmd(params))
    assert res == 0
    out, err = capfd.readouterr()
    assert str(tmp_path) in out.split(os.linesep)


def test_with_path_result_code(tmp_path):
    """
    result code should be non-zero on error
    """
    params = ['-c', "raise ValueError()"]
    res = launch.with_path(tmp_path, launch.infer_cmd(params))
    assert res > 0


@pytest.fixture
def clean_pythonpath(monkeypatch):
    monkeypatch.delitem(os.environ, 'PYTHONPATH', raising=False)


def test_build_env(clean_pythonpath):
    os.environ['PYTHONPATH'] = 'something'
    env = launch._build_env(pathlib.Path('else'))
    expected = os.pathsep.join(('else', 'something'))
    assert env['PYTHONPATH'] == expected

    os.environ['PYTHONPATH'] = ''
    env = launch._build_env(pathlib.Path('something'))
    assert env['PYTHONPATH'] == 'something'

    initial = os.pathsep.join(['something', 'else'])
    os.environ['PYTHONPATH'] = initial
    env = launch._build_env(pathlib.Path('a'))
    expected = os.pathsep.join(['a', 'something', 'else'])
    assert env['PYTHONPATH'] == expected



#!/usr/bin/env pip-run
import cowsay
import requests
from bs4 import BeautifulSoup as BS

res = requests.get('https://python.org')
b = BS(res.text, 'html.parser')
cowsay.dragon(b.find("div", class_="introduction").get_text())
